@@ -3,6 +3,8 @@ import logging
 import socket
 import errno
 import select
+import sys
+from six.moves import xrange
 from .collate import *
 from .tdsproto import *
 
@@ -28,6 +30,7 @@ TDS71 = 0x71000000
 TDS71rev1 = 0x71000001
 TDS72 = 0x72090002
 TDS73A = 0x730A0003
+TDS73 = TDS73A
 TDS73B = 0x730B0003
 TDS74 = 0x74000004
 
@@ -43,17 +46,17 @@ TDSSELERR = 0
 TDSPOLLURG = 0x8000
 
 
-IS_TDS42 = lambda(x): x.tds_version == 0x402
-IS_TDS46 = lambda(x): x.tds_version == 0x406
-IS_TDS50 = lambda(x): x.tds_version == 0x500
-IS_TDS70 = lambda(x): x.tds_version == TDS70
-IS_TDS71 = lambda(x): x.tds_version in (TDS71, TDS71rev1)
-IS_TDS72 = lambda(x): x.tds_version == TDS72
-IS_TDS73 = lambda(x): x.tds_version in (TDS73A, TDS73B)
-IS_TDS7_PLUS = lambda(x): x.tds_version >= TDS70
-IS_TDS71_PLUS = lambda(x): x.tds_version >= TDS71
-IS_TDS72_PLUS = lambda(x): x.tds_version >= TDS72
-IS_TDS73_PLUS = lambda(x): x.tds_version >= TDS73A
+IS_TDS42 = lambda x: x.tds_version == 0x402
+IS_TDS46 = lambda x: x.tds_version == 0x406
+IS_TDS50 = lambda x: x.tds_version == 0x500
+IS_TDS70 = lambda x: x.tds_version == TDS70
+IS_TDS71 = lambda x: x.tds_version in (TDS71, TDS71rev1)
+IS_TDS72 = lambda x: x.tds_version == TDS72
+IS_TDS73 = lambda x: x.tds_version in (TDS73A, TDS73B)
+IS_TDS7_PLUS = lambda x: x.tds_version >= TDS70
+IS_TDS71_PLUS = lambda x: x.tds_version >= TDS71
+IS_TDS72_PLUS = lambda x: x.tds_version >= TDS72
+IS_TDS73_PLUS = lambda x: x.tds_version >= TDS73A
 
 TDS_NO_COUNT = -1
 
@@ -87,12 +90,13 @@ TDS_HANDLE_ALL = 0
 
 
 def _gen_return_flags():
-    _globs = globals()
+    _globs = {}
     prefix = 'TDS_TOKEN_RES_'
     for key, value in globals().items():
         if key.startswith(prefix):
             _globs['TDS_RETURN_' + key[len(prefix):]] = 1 << (value * 2)
             _globs['TDS_STOPAT_' + key[len(prefix):]] = 2 << (value * 2)
+    globals().update(_globs)
 _gen_return_flags()
 
 
@@ -123,23 +127,23 @@ TDS_SUCCESS = 0
 TDS_FAIL = -1
 TDS_CANCELLED = -2
 
-TDS_FAILED = lambda(rc): rc < 0
-TDS_SUCCEED = lambda(rc): rc >= 0
+TDS_FAILED = lambda rc: rc < 0
+TDS_SUCCEED = lambda rc: rc >= 0
 
-is_blob_type = lambda(x): x in (SYBTEXT, SYBIMAGE, SYBNTEXT)
-is_blob_col = lambda(col): (col.column_varint_size > 2)
+is_blob_type = lambda x: x in (SYBTEXT, SYBIMAGE, SYBNTEXT)
+is_blob_col = lambda col: (col.column_varint_size > 2)
 # large type means it has a two byte size field
 # define is_large_type(x) (x>128)
-is_numeric_type = lambda(x): x in (SYBNUMERIC, SYBDECIMAL)
-is_unicode_type = lambda(x): x in (XSYBNVARCHAR, XSYBNCHAR, SYBNTEXT, SYBMSXML)
-is_collate_type = lambda(x): x in (XSYBVARCHAR, XSYBCHAR, SYBTEXT, XSYBNVARCHAR, XSYBNCHAR, SYBNTEXT)
-is_ascii_type = lambda(x): x in (XSYBCHAR, XSYBVARCHAR, SYBTEXT, SYBCHAR, SYBVARCHAR)
-is_char_type = lambda(x): is_unicode_type(x) or is_ascii_type(x)
-is_similar_type = lambda(x, y): is_char_type(x) and is_char_type(y) or is_unicode_type(x) and is_unicode_type(y)
+is_numeric_type = lambda x: x in (SYBNUMERIC, SYBDECIMAL)
+is_unicode_type = lambda x: x in (XSYBNVARCHAR, XSYBNCHAR, SYBNTEXT, SYBMSXML)
+is_collate_type = lambda x: x in (XSYBVARCHAR, XSYBCHAR, SYBTEXT, XSYBNVARCHAR, XSYBNCHAR, SYBNTEXT)
+is_ascii_type = lambda x: x in (XSYBCHAR, XSYBVARCHAR, SYBTEXT, SYBCHAR, SYBVARCHAR)
+is_char_type = lambda x: is_unicode_type(x) or is_ascii_type(x)
+is_similar_type = lambda x, y: is_char_type(x) and is_char_type(y) or is_unicode_type(x) and is_unicode_type(y)
 
-tds_conn = lambda(tds): tds
+tds_conn = lambda tds: tds
 
-IS_TDSDEAD = lambda(tds): tds is None or tds._sock is None
+IS_TDSDEAD = lambda tds: tds is None or tds._sock is None
 
 TDS_DEF_BLKSZ = 512
 TDS_DEF_CHARSET = "iso_1"
@@ -181,9 +185,9 @@ def tds_quote_id(tds, id):
 
 
 # Check if product is Sybase (such as Adaptive Server Enterrprice). x should be a TDSSOCKET*.
-TDS_IS_SYBASE = lambda(x): not tds_conn(x).product_version & 0x80000000
+TDS_IS_SYBASE = lambda x: not tds_conn(x).product_version & 0x80000000
 # Check if product is Microsft SQL Server. x should be a TDSSOCKET*.
-TDS_IS_MSSQL = lambda(x): tds_conn(x).product_version & 0x80000000
+TDS_IS_MSSQL = lambda x: tds_conn(x).product_version & 0x80000000
 
 # store a tuple of programming error codes
 prog_errors = (
@@ -203,16 +207,23 @@ integrity_errors = (
 )
 
 
+if sys.version_info[0] >= 3:
+    exc_base_class = Exception
+else:
+    exc_base_class = StandardError
+
+
 # exception hierarchy
-class Warning(StandardError):
+class Warning(exc_base_class):
     pass
 
 
-class Error(StandardError):
+class Error(exc_base_class):
     pass
 
 
-TimeoutError = socket.timeout
+class TimeoutError(Error):
+    pass
 
 
 class InterfaceError(Error):
@@ -233,11 +244,6 @@ class DatabaseError(Error):
                 self.state, self.line, self.text)
 
 
-class ClosedConnectionError(InterfaceError):
-    def __init__(self):
-        super(ClosedConnectionError, self).__init__('Server closed connection')
-
-
 class DataError(Error):
     pass
 
@@ -268,12 +274,31 @@ class NotSupportedError(DatabaseError):
 #############################
 ## DB-API type definitions ##
 #############################
-STRING = 1
-BINARY = 2
-NUMBER = 3
-DATETIME = 4
-DECIMAL = 5
-ROWID = 6
+class DBAPITypeObject:
+    def __init__(self, *values):
+        self.values = set(values)
+
+    def __eq__(self, other):
+        return other in self.values
+
+    def __cmp__(self, other):
+        if other in self.values:
+            return 0
+        if other < self.values:
+            return 1
+        else:
+            return -1
+
+STRING = DBAPITypeObject(SYBVARCHAR, SYBCHAR, SYBTEXT,
+                         XSYBNVARCHAR, XSYBNCHAR, SYBNTEXT,
+                         XSYBVARCHAR, XSYBCHAR, SYBMSXML)
+BINARY = DBAPITypeObject(SYBIMAGE, SYBBINARY, SYBVARBINARY, XSYBVARBINARY, XSYBBINARY)
+NUMBER = DBAPITypeObject(SYBBIT, SYBINT1, SYBINT2, SYBINT4, SYBINT8, SYBINTN,
+                         SYBREAL, SYBFLT8, SYBFLTN)
+DATETIME = DBAPITypeObject(SYBDATETIME, SYBDATETIME4, SYBDATETIMN)
+DECIMAL = DBAPITypeObject(SYBMONEY, SYBMONEY4, SYBMONEYN, SYBNUMERIC,
+                          SYBDECIMAL)
+ROWID = DBAPITypeObject()
 
 
 # stored procedure output parameter
@@ -297,7 +322,7 @@ class output:
         self._value = value
 
 
-class Binary(str):
+class Binary(bytes):
     def __repr__(self):
         return 'Binary({0})'.format(super(Binary, self).__repr__())
 
@@ -396,12 +421,12 @@ def skipall(stm, size):
     if len(res) == size:
         return
     elif len(res) == 0:
-        raise ClosedConnectionError()
+        raise Error('Server closed connection')
     left = size - len(res)
     while left:
         buf = stm.read(left)
         if len(buf) == 0:
-            raise ClosedConnectionError()
+            raise Error('Server closed connection')
         left -= len(buf)
 
 
@@ -410,17 +435,26 @@ def readall(stm, size):
     if len(res) == size:
         return res
     elif len(res) == 0:
-        raise ClosedConnectionError()
+        raise Error('Server closed connection')
     chunks = [res]
     left = size - len(res)
     while left:
         buf = stm.read(left)
         if len(buf) == 0:
-            raise ClosedConnectionError()
+            raise Error('Server closed connection')
         chunks.append(buf)
         left -= len(buf)
     return b''.join(chunks)
 
+def readall_fast(stm, size):
+    buf, offset = stm.read_fast(size)
+    if len(buf) - offset < size:
+        # slow case
+        buf = buf[offset:]
+        buf += stm.read(size - len(buf))
+        return buf, 0
+    return buf, offset
+
 
 class _TdsReader(object):
     def __init__(self, session, emul_little_endian):
@@ -438,8 +472,21 @@ class _TdsReader(object):
     def packet_type(self):
         return self._type
 
+    def read_fast(self, size):
+        if self._pos >= len(self._buf):
+            if self._have >= self._size:
+                self._read_packet()
+            else:
+                self._buf = self._transport.read(self._size - self._have)
+                self._pos = 0
+                self._have += len(self._buf)
+        offset = self._pos
+        self._pos += size
+        return self._buf, offset
+
     def unpack(self, struct):
-        return struct.unpack(readall(self, struct.size))
+        buf, offset = readall_fast(self, struct.size)
+        return struct.unpack_from(buf, offset)
 
     def get_byte(self):
         return self.unpack(_byte)[0]
@@ -505,16 +552,8 @@ class _TdsReader(object):
             left -= len(buf)
 
     def read(self, size):
-        if self._pos >= len(self._buf):
-            if self._have >= self._size:
-                self._read_packet()
-            else:
-                self._buf = self._transport.read(self._size - self._have)
-                self._pos = 0
-                self._have += len(self._buf)
-        res = self._buf[self._pos:self._pos + size]
-        self._pos += len(res)
-        return res
+        buf, offset = self.read_fast(size)
+        return buf[offset:offset + size]
 
     def _read_packet(self):
         try:
@@ -551,7 +590,7 @@ class _TdsWriter(object):
             return
 
         if bufsize > len(self._buf):
-            self._buf.extend('\0' * (bufsize - len(self._buf)))
+            self._buf.extend(b'\0' * (bufsize - len(self._buf)))
         else:
             self._buf = self._buf[0:bufsize]
 
@@ -656,6 +695,19 @@ class MemoryChunkedHandler(object):
         #logger.debug('MemoryChunkedHandler.new_chunk(sz=%d)', len(val))
         self._chunks.append(val)
 
+    def end(self):
+        return b''.join(self._chunks)
+
+
+class MemoryStrChunkedHandler(object):
+    def begin(self, column, size):
+        self.size = size
+        self._chunks = []
+
+    def new_chunk(self, val):
+        #logger.debug('MemoryChunkedHandler.new_chunk(sz=%d)', len(val))
+        self._chunks.append(val)
+
     def end(self):
         return ''.join(self._chunks)
 
@@ -710,16 +762,16 @@ class _TdsSession(object):
         if state == TDS_PENDING:
             if prior_state in (TDS_READING, TDS_QUERYING):
                 self.state = TDS_PENDING
-                tds_mutex_unlock(self.wire_mtx)
+                #tds_mutex_unlock(self.wire_mtx)
             else:
                 logger.error('logic error: cannot chage query state from {0} to {1}'.
                              format(state_names[prior_state], state_names[state]))
         elif state == TDS_READING:
             # transition to READING are valid only from PENDING
-            if tds_mutex_trylock(self.wire_mtx):
-                return self.state
+            #if tds_mutex_trylock(self.wire_mtx):
+            #    return self.state
             if self.state != TDS_PENDING:
-                tds_mutex_unlock(self.wire_mtx)
+                #tds_mutex_unlock(self.wire_mtx)
                 logger.error('logic error: cannot change query state from {0} to {1}'.
                              format(state_names[prior_state], state_names[state]))
             else:
@@ -728,24 +780,24 @@ class _TdsSession(object):
             if prior_state == TDS_DEAD:
                 logger.error('logic error: cannot change query state from {0} to {1}'.
                              format(state_names[prior_state], state_names[state]))
-            elif prior_state in (TDS_READING, TDS_QUERYING):
-                tds_mutex_unlock(self.wire_mtx)
+            #elif prior_state in (TDS_READING, TDS_QUERYING):
+            #    tds_mutex_unlock(self.wire_mtx)
             self.state = state
         elif state == TDS_DEAD:
-            if prior_state in (TDS_READING, TDS_QUERYING):
-                tds_mutex_unlock(self.wire_mtx)
+            #if prior_state in (TDS_READING, TDS_QUERYING):
+            #    tds_mutex_unlock(self.wire_mtx)
             self.state = state
         elif state == TDS_QUERYING:
-            if tds_mutex_trylock(self.wire_mtx):
-                return self.state
+            #if tds_mutex_trylock(self.wire_mtx):
+            #    return self.state
             if self.state == TDS_DEAD:
-                tds_mutex_unlock(self.wire_mtx)
-                logger.error('logic error: cannot change query state from {0} to {1}'.
-                             format(state_names[prior_state], state_names[state]))
+                #tds_mutex_unlock(self.wire_mtx)
+                raise InterfaceError('logic error: cannot change query state from {0} to {1}'.
+                                     format(state_names[prior_state], state_names[state]))
             elif self.state != TDS_IDLE:
-                tds_mutex_unlock(self.wire_mtx)
-                logger.error('logic error: cannot change query state from {0} to {1}'.
-                             format(state_names[prior_state], state_names[state]))
+                #tds_mutex_unlock(self.wire_mtx)
+                raise InterfaceError('logic error: cannot change query state from {0} to {1}'.
+                                     format(state_names[prior_state], state_names[state]))
             else:
                 self.rows_affected = TDS_NO_COUNT
                 self.internal_sp_called = 0
@@ -775,7 +827,7 @@ class _StateContext(object):
 
 
 class _TdsSocket(object):
-    def __init__(self, login, sock):
+    def __init__(self, login):
         self._is_connected = False
         self._bufsize = login.blocksize
         self.login = None
@@ -794,32 +846,67 @@ class _TdsSocket(object):
 
         # Jeff's hack, init to no timeout
         self.query_timeout = login.connect_timeout if login.connect_timeout else login.query_timeout
-        self._sock = sock
+        self._sock = None
         import socket
         if hasattr(socket, 'socketpair'):
             tds_conn(self).s_signal, tds_conn(self).s_signaled = socket.socketpair(socket.AF_UNIX, socket.SOCK_DGRAM)
-        self.tds_version = login.tds_version
-        self.emul_little_endian = login.emul_little_endian
-        if IS_TDS7_PLUS(self):
-            # TDS 7/8 only supports little endian
-            self.emul_little_endian = True
-        from login import tds_login
-        tds_login(self._main_session, login)
-        text_size = login.text_size
-        if self.mars_enabled:
-            self._setup_smp()
-        self._is_connected = True
-        q = []
-        if text_size:
-            q.append('set textsize {0}'.format(int(text_size)))
-        if login.database and self.env.database != login.database:
-            q.append('use ' + tds_quote_id(self, login.database))
-        if q:
-            tds_submit_query(tds._main_session, ''.join(q))
-            tds_process_simple_query(tds._main_session)
+        try:
+            self.tds_version = login.tds_version
+            self.emul_little_endian = login.emul_little_endian
+            if IS_TDS7_PLUS(self):
+                # TDS 7/8 only supports little endian
+                self.emul_little_endian = True
+            if IS_TDS7_PLUS(self) and login.instance_name and not login.port:
+                instances = tds7_get_instances(login.server_name)
+                if login.instance_name not in instances:
+                    raise LoginError("Instance {0} not found on server {1}".format(login.instance_name, login.server_name))
+                instdict = instances[login.instance_name]
+                if 'tcp' not in instdict:
+                    raise LoginError("Instance {0} doen't have tcp connections enabled".format(login.instance_name))
+                login.port = int(instdict['tcp'])
+            connect_timeout = login.connect_timeout
+
+            if not login.port:
+                login.port = 1433
+            err = None
+            for host in login.load_balancer.choose():
+                try:
+                    tds_open_socket(self, host, login.port, connect_timeout)
+                except socket.error as e:
+                    err = LoginError("Cannot connect to server '{0}': {1}".format(host, e), e)
+                    continue
+                try:
+                    from .login import tds_login
+                    tds_login(self._main_session, login)
+                    text_size = login.text_size
+                    if self.mars_enabled:
+                        self._setup_smp()
+                    self._is_connected = True
+                    q = []
+                    if text_size:
+                        q.append('set textsize {0}'.format(int(text_size)))
+                    if login.database and self.env.database != login.database:
+                        q.append('use ' + tds_quote_id(self, login.database))
+                    if q:
+                        tds_submit_query(tds._main_session, ''.join(q))
+                        tds_process_simple_query(tds._main_session)
+                except Exception as e:
+                    self._sock.close()
+                    err = e
+                    #raise
+                    continue
+                break
+            else:
+                raise err
+        except Exception:
+            if tds_conn(self).s_signal is not None:
+                tds_conn(self).s_signal.close()
+            if tds_conn(self).s_signaled is not None:
+                tds_conn(self).s_signaled.close()
+            raise
 
     def _setup_smp(self):
-        from smp import SmpManager
+        from .smp import SmpManager
         self._smp_manager = SmpManager(self)
         self._main_session = _TdsSession(self, self._smp_manager.create_session())
 
@@ -836,10 +923,19 @@ class _TdsSocket(object):
 
     def read(self, size):
         try:
-            buf = self._sock.recv(size)
-            if len(buf) == 0:
-                raise ClosedConnectionError()
-            return buf
+            events = tds_select(self, TDSSELREAD, self.query_timeout)
+            if events & TDSPOLLURG:
+                buf = tds_conn(self).s_signaled.read(size)
+                if not self.in_cancel:
+                    tds_put_cancel(self)
+                return buf
+            elif events:
+                buf = self._sock.recv(size)
+                if len(buf) == 0:
+                    raise Error('Server closed connection')
+                return buf
+            else:
+                raise TimeoutError('Timeout')
         except TimeoutError:
             raise
         except:
@@ -853,6 +949,9 @@ class _TdsSocket(object):
         try:
             pos = 0
             while pos < len(data):
+                res = tds_select(self, TDSSELWRITE, self.query_timeout)
+                if not res:
+                    raise TimeoutError('Timeout')
                 flags = 0
                 if hasattr(socket, 'MSG_NOSIGNAL'):
                     flags |= socket.MSG_NOSIGNAL
@@ -934,8 +1033,8 @@ def tds_select(tds, tds_sel, timeout_seconds):
     poll_seconds = 1 if tds.int_handler else timeout_seconds
     seconds = timeout_seconds
     while timeout_seconds is None or seconds > 0:
-        timeout = poll_seconds * 1000 if poll_seconds else None
         if USE_POLL:
+            timeout = poll_seconds * 1000 if poll_seconds else None
             poll = select.poll()
             poll.register(tds._sock, tds_sel)
             poll.register(tds_conn(tds).s_signaled, select.POLLIN)
@@ -953,6 +1052,7 @@ def tds_select(tds, tds_sel, timeout_seconds):
             if tds.int_handler:
                 tds.int_handler()
         else:
+            timeout = poll_seconds if poll_seconds else None
             read = []
             write = []
             if tds_sel == TDSSELREAD:
